import sys

PY = sys.version_info[:2]
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3
PYPY = '__pypy__' in sys.builtin_module_names


stringtypes = (bytes, str, )
integertypes = (int, )
unicodestringtype = str
bytestringtype = bytes

INT2BYTE_CACHE = {i:bytes([i]) for i in range(256)}
def int2byte(character):
    """Converts (0 through 255) integer into b'...' character."""
    return INT2BYTE_CACHE[character]

BYTE2INT_CACHE = {bytes([i]):i for i in range(256)}
def byte2int(character):
    """Converts b'...' character into (0 through 255) integer."""
    return BYTE2INT_CACHE[character]

def str2bytes(string):
    """Converts '...' string into b'...' string. On PY2 they are equivalent. On PY3 its utf8 encoded."""
    return string.encode("utf8")

def bytes2str(string):
    """Converts b'...' string into '...' string. On PY2 they are equivalent. On PY3 its utf8 decoded."""
    return string.decode("utf8")

ITERATEBYTES_CACHE = {i:bytes([i]) for i in range(256)}
def iteratebytes(data):
    """Iterates though b'...' string yielding b'...' characters."""
    return (ITERATEBYTES_CACHE[i] for i in data)

def iterateints(data):
    """Iterates though b'...' string yielding (0 through 255) integers."""
    return data

def reprstring(data):
    """Ensures there is b- u- prefix before the string."""
    if isinstance(data, bytes):
        return repr(data)
    if isinstance(data, str):
        return 'u' + repr(data)

def trimstring(data):
    """Trims b- u- prefix"""
    if isinstance(data, bytes):
        return repr(data)[1:]
    if isinstance(data, str):
        return repr(data)

def integers2bytes(ints):
    """Converts integer generator into bytes."""
    return bytes(ints)

def bytes2integers(data):
    """Converts bytes into bytes/bytearray, so indexing/iterating yields integers."""
    return data



"""
what : snoop v2 capture file.
 how : http://tools.ietf.org/html/rfc1761
 who : jesse@housejunkie.ca
"""

from construct import *
import time


# use core Timestamp
class EpochTimeStampAdapter(Adapter):
    """ Convert epoch timestamp <-> localtime """
    def _decode(self, obj, context, path):
        return time.ctime(obj)
    def _encode(self, obj, context, path):
        return int(time.mktime(time.strptime(obj)))

packet_record = Struct(
    "original_length" / Int32ub,
    "included_length" / Int32ub,
    "record_length" / Int32ub,
    "cumulative_drops" / Int32ub,
    "timestamp_seconds" / EpochTimeStampAdapter(Int32ub),
    "timestamp_microseconds" / Int32ub,
    "data" / Bytes(this.included_length),
    # 24 being the static length of the packet_record header
    Padding(this.record_length - this.included_length - 24),
)

datalink_type = Enum(Int32ub,
    IEEE802dot3 = 0,
    IEEE802dot4 = 1,
    IEEE802dot5 = 2,
    IEEE802dot6 = 3,
    ETHERNET = 4,
    HDLC = 5,
    CHARSYNC = 6,
    IBMCHANNEL = 7,
    FDDI = 8,
    OTHER = 9,
    UNASSIGNED = 10,
)

snoop_file = Struct(
    "signature" / Const(b"snoop\x00\x00\x00"),
    "version" / Int32ub, # snoop v1 is deprecated
    "datalink" / datalink_type,
    "records" / GreedyRange(packet_record),
)



from declarativeunittest import *
from construct import *
from construct.lib import *

from gallery import pe32file


def test_pe32():
    commondump(pe32file, "python37-win32.exe")
    commondump(pe32file, "python37-win64.exe")
    commondump(pe32file, "SharpZipLib0860-dotnet20.dll")
    commondump(pe32file, "sqlite3.dll")



"""
Windows/OS2 Bitmap (BMP) this could have been a perfect show-case file format, but they had to make it ugly (all sorts of alignments)
"""

from construct import *

#===============================================================================
# pixels: uncompressed
#===============================================================================
def UncompressedRows(subcon, align_to_byte=False):
    """argh! lines must be aligned to a 4-byte boundary, and bit-pixel lines must be aligned to full bytes..."""
    if align_to_byte:
        line_pixels = Bitwise(Aligned(8, Array(this.width, subcon)))
    else:
        line_pixels = Array(this.width, subcon)
    return Array(this.height, Aligned(4, line_pixels))

uncompressed_pixels = Switch(this.bpp,
    {
        1 : UncompressedRows(Bit, align_to_byte=True), # index
        4 : UncompressedRows(Nibble, align_to_byte=True), # index
        8 : UncompressedRows(Byte),  # index
        24 : UncompressedRows(Byte[3]),  # rgb
    }
)

#===============================================================================
# pixels: Run Length Encoding (RLE) 8 bit
#===============================================================================
class RunLengthAdapter(Adapter):
    def _decode(self, obj, context, path):
        length,value = obj
        return [value] * length
    def _encode(self, obj, context, path):
        return len(obj), obj[0]

# WARNING: not used anywhere
rle8pixel = "rle8pixel" / RunLengthAdapter(Byte >> Byte)

#===============================================================================
# file structure
#===============================================================================
bitmap_file = Struct(
    "signature" / Const(b"BM"),
    "file_size" / Int32ul,
    Padding(4),
    "data_offset" / Int32ul,
    "header_size" / Int32ul,
    "version" / Enum(Computed(this.header_size),
        v2 = 12,
        v3 = 40,
        v4 = 108,
    ),
    "width" / Int32ul,
    "height" / Int32ul,
    "number_of_pixels" / Computed(this.width * this.height),
    "planes" / Int16ul,
    "bpp" / Int16ul, # bits per pixel
    "compression" / Enum(Int32ul,
        Uncompressed = 0,
        RLE8 = 1,
        RLE4 = 2,
        Bitfields = 3,
        JPEG = 4,
        PNG = 5,
    ),
    "image_data_size" / Int32ul, # in bytes
    "horizontal_dpi" / Int32ul,
    "vertical_dpi" / Int32ul,
    "colors_used" / Int32ul,
    "important_colors" / Int32ul,

    # palette (24 bit has no palette)
    # NOTE: was called "rgb" inside of it
    "palette" / Array(lambda ctx: 2**ctx.bpp if ctx.bpp <= 8 else 0, 
        Padded(4, Byte[3])),

    "pixels" / Pointer(this.data_offset,
        Switch(this.compression, {"Uncompressed" : uncompressed_pixels}),
    ),
)



# -*- coding: utf-8 -*-

from declarativeunittest import *
from construct import *
from construct.lib import *


example = Struct(
    "num" / Byte,

    "bytes1" / Bytes(4),
    "bytes2" / Bytes(this.num),
    "greedybytes" / Prefixed(Byte, GreedyBytes),
    "bitwise1" / Bitwise(BitsInteger(8, swapped=False)),
    "bitwise2" / Bitwise(BitsInteger(8, swapped=True)),
    "bytewise1" / Bitwise(Bytewise(BytesInteger(16, swapped=True))),
    "bytewise2" / Bitwise(Bytewise(BytesInteger(16, swapped=False))),

    "formatfield" / FormatField(">", "B"),
    "bytesinteger0" / BytesInteger(0),
    "bytesinteger1" / BytesInteger(16, swapped=True),
    "bytesinteger2" / BytesInteger(16, swapped=False),
    "bytesinteger3" / BytesInteger(this.num),
    "bitsinteger0" / Bitwise(BitsInteger(0)),
    "bitsinteger1" / Bitwise(BitsInteger(8, swapped=False)),
    "bitsinteger2" / Bitwise(BitsInteger(8, swapped=True)),
    # - unknown cause (SizeofError: cannot calculate size, key not found in context)
    # "bitsinteger3" / Bitwise(BitsInteger(this.num)),
    "varint" / VarInt,
    "byte" / Byte,
    "float1" / Single,
    "float2" / Double,

    "string2" / PaddedString(12, "ascii"),
    "string2" / PaddedString(12, "utf8"),
    "string3" / PaddedString(12, "utf16"),
    "string4" / PaddedString(12, "utf32"),
    "pascalstring1" / PascalString(Byte, "ascii"),
    "pascalstring2" / PascalString(Byte, "utf8"),
    "cstring1" / CString("ascii"),
    "cstring2" / CString("utf8"),
    "cstring3" / CString("utf16"),
    "cstring4" / CString("utf32"),
    "greedystring1" / Prefixed(Byte, GreedyString("ascii")),
    "greedystring2" / Prefixed(Byte, GreedyString("utf8")),

    "flag" / Flag,
    "enum1" / Enum(Byte, zero=0),
    "enum2" / Enum(Byte),
    "flagsenum1" / FlagsEnum(Byte, zero=0, one=1),
    "flagsenum2" / FlagsEnum(Byte),
    "mapping" / Mapping(Byte, {"zero":0}),

    "struct" / Struct("field" / Byte),
    "sequence1" / Sequence(Byte, Byte),
    "sequence2" / Sequence("num1" / Byte, "num2" / Byte),

    "array1" / Array(5, Byte),
    "array2" / Array(this.num, Byte),
    "greedyrange0" / Prefixed(Byte, GreedyRange(Byte)),
    "repeatuntil1" / RepeatUntil(obj_ == 0, Byte),

    "const1" / Const(bytes(4)),
    "const2" / Const(0, Int32ub),
    "computed" / Computed(this.num),
    "index1" / Array(3, Index),
    "index2" / RestreamData(b"\x00", GreedyRange(Byte >> Index)),
    "index3" / RestreamData(b"\x00", RepeatUntil(True, Byte >> Index)),
    "rebuild" / Rebuild(Byte, len_(this.array1)),
    "default" / Default(Byte, 0),
    Check(this.num == 0),
    "check" / Check(this.num == 0),
    "error0" / If(False, Error),
    "focusedseq1" / FocusedSeq("num", Const(bytes(4)), "num"/Byte),
    "focusedseq2_select" / Computed("num"),
    "focusedseq2" / FocusedSeq(this._.focusedseq2_select, "num"/Byte),
    "pickled_data" / Computed(b"(lp0\n(taI1\naF2.3\na(dp1\na(lp2\naS'1'\np3\naS''\np4\na."),
    "pickled" / RestreamData(this.pickled_data, Pickled),
    "numpy_data" / Computed(b"\x93NUMPY\x01\x00F\x00{'descr': '<i8', 'fortran_order': False, 'shape': (3,), }            \n\x01\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00"),
    "numpy1" / RestreamData(this.numpy_data, Numpy),
    "namedtuple1" / NamedTuple("coord", "x y z", Byte[3]),
    "namedtuple2" / RestreamData(b"\x00\x00\x00", NamedTuple("coord", "x y z", GreedyRange(Byte))),
    "namedtuple3" / NamedTuple("coord", "x y z", Byte >> Byte >> Byte),
    "namedtuple4" / NamedTuple("coord", "x y z", "x"/Byte + "y"/Byte + "z"/Byte),
    "timestamp1" / RestreamData(b'\x00\x00\x00\x00ZIz\x00', Timestamp(Int64ub, 1, 1970)),
    "timestamp2" / RestreamData(b'H9\x8c"', Timestamp(Int32ub, "msdos", "msdos")),
    "hex1" / Hex(Byte),
    "hex2" / Hex(Bytes(1)),
    "hex3" / Hex(RawCopy(Byte)),
    "hexdump1" / HexDump(Bytes(1)),
    "hexdump2" / HexDump(RawCopy(Byte)),

    "union1" / Union(None, "char"/Byte, "short"/Short, "int"/Int),
    "union2" / Union(1, "char"/Byte, "short"/Short, "int"/Int),
    "union3" / Union(0, "char1"/Byte, "char2"/Byte, "char3"/Byte),
    "union4" / Union("char1", "char1"/Byte, "char2"/Byte, "char3"/Byte),
    "select" / Select(Byte, CString("ascii")),
    "optional" / Optional(Byte),
    "if1" / If(this.num == 0, Byte),
    "ifthenelse" / IfThenElse(this.num == 0, Byte, Byte),
    "switch1" / Switch(this.num, {0 : Byte, 255 : Error}),
    "switch2" / Switch(this.num, {}),
    "switch3" / Switch(this.num, {}, default=Byte),
    "stopif0" / StopIf(this.num == 255),
    "stopif1" / Struct(StopIf(this._.num == 0), Error),
    "stopif2" / Sequence(StopIf(this._.num == 0), Error),
    "stopif3" / GreedyRange(StopIf(this.num == 0)),

    "padding" / Padding(2),
    "paddedbyte" / Padded(4, Byte),
    "alignedbyte" / Aligned(4, Byte),
    "alignedstruct" / AlignedStruct(4, "a"/Byte, "b"/Short),
    "bitstruct" / BitStruct("a"/Octet),

    "pointer" / Pointer(0, Byte),
    "peek" / Peek(Byte),
    "seek0" / Seek(0, 1),
    "tell" / Tell,
    "pass1" / Pass,
    "terminated0" / Prefixed(Byte, Terminated),

    "rawcopy1" / RawCopy(Byte),
    "rawcopy2" / RawCopy(RawCopy(RawCopy(Byte))),
    "bytesswapped" / ByteSwapped(BytesInteger(8)),
    "bitsswapped" / BitsSwapped(BytesInteger(8)),
    "prefixed1" / Prefixed(Byte, GreedyBytes),
    "prefixed2" / RestreamData(b"\x01", Prefixed(Byte, GreedyBytes, includelength=True)),
    "prefixedarray" / PrefixedArray(Byte, Byte),
    "fixedsized" / FixedSized(10, GreedyBytes),
    "nullterminated" / RestreamData(b'\x01\x00', NullTerminated(GreedyBytes)),
    "nullstripped" / RestreamData(b'\x01\x00', NullStripped(GreedyBytes)),
    "restreamdata" / RestreamData(b"\xff", Byte),
    "restreamdata_verify" / Check(this.restreamdata == 255),
    # Transformed
    # Restreamed
    # ProcessXor
    # ProcessRotateLeft
    # Checksum
    "compressed_bzip2_data" / Computed(b'BZh91AY&SYSc\x11\x99\x00\x00\x00A\x00@\x00@\x00 \x00!\x00\x82\x83\x17rE8P\x90Sc\x11\x99'),
    "compressed_bzip2" / RestreamData(this.compressed_bzip2_data, Compressed(GreedyBytes, "bzip2", level=9)),
    # Rebuffered

    # Lazy
    # LazyStruct
    # LazyArray
    # LazyBound

    # adapters and validators

    Probe(),
    # - fails due to unknown causes (Expected an identifier or literal)
    # "probe" / Probe(),
    "debugger" / Debugger(Byte),

    "items1" / Computed([1,2,3]),
    "len_" / Computed(len_(this.items1)),
    # - faulty list_ implementation, compiles into correct code
    # "repeatuntil2" / RepeatUntil(list_ == [0], Byte),
    # "repeatuntil3" / RepeatUntil(list_[-1] == 0, Byte),
)
exampledata = bytes(1000)


def test_example_benchmark():
    d = example.compile(filename="example_compiled.py")
    d.benchmark(exampledata, filename="example_benchmark.txt")

def test_compiled_benchmark():
    d = Struct().compile()
    d.benchmark(b"")
    d = Struct(Error).compile()
    d.benchmark(b"")

def test_compiler_recursion():
    raises(Construct().compile) == NotImplementedError



"""
fat.py; ad-hoc fat16 reader
    by Bram Westerbaan <bram@westerbaan.name> 

references:
    http://en.wikipedia.org/wiki/File_Allocation_Table
    http://www.ecma-international.org/publications/standards/Ecma-107.htm

example:
    with open("/dev/sdc1", "rb") as file:
        fs = FatFs(file)
        for rootdir in fs:
            print rootdir
"""

import numbers
from io import BytesIO, BufferedReader
from construct import *

Fat16Header = Struct(
    "jumpInstruction" / Bytes(3),
    "creatingSystemId" / Bytes(8),
    "sectorSize" / Int16ul,
    "sectorsPerCluster" / Byte,
    "reservedSectorCount" / Int16ul,
    "fatCount" / Byte,
    "rootdirEntryCount" / Int16ul,
    "sectorCount_small" / Int16ul,
    "mediaId" / Byte, 
    "sectorsPerFat" / Int16ul,
    "sectorsPerTrack" / Int16ul,
    "sideCount" / Int16ul,
    "hiddenSectorCount" / Int32ul,
    "sectorCount_large" / Int32ul,
    "physicalDriveNumber" / Byte,
    "currentHead" / Byte,
    "extendedBootSignature" / Byte,
    "volumeId" / Bytes(4),
    "volumeLabel" / Bytes(11),
    "fsType" / Const(b"FAT16   "),
    "bootCode" / Bytes(448),
    "bootSectorSignature" / Const(b"\x55\xaa"),
)

BootSector = Struct(
    Embedded(Fat16Header),
    Padding(this.sectorSize - Fat16Header.sizeof()),
)

FatEntry = Enum(Int16ul,
    free_cluster = 0x0000,
    bad_cluster = 0xfff7,
    last_cluster = 0xffff,
    default = Pass,
)

DirEntry = Struct(
    "name" / Bytes(8),
    "extension" / Bytes(3),
    "attributes" / BitStruct(
        "unused" / Flag,
        "device" / Flag,
        "archive" / Flag,
        "subDirectory" / Flag,
        "volumeLabel" / Flag,
        "system" /Flag,
        "hidden" / Flag,
        "readonly" / Flag,
    ),
    Padding(10),
    "timeRecorded" / Int16ul,
    "dateRecorded" / Int16ul,
    "firstCluster" / Int16ul,
    "fileSize" / Int32ul,
)

# NOTE: non-integer division, really?
PreDataRegion = Struct(
    "bootSector" / Embedded(BootSector),
    # the remaining reserved sectors
    Padding((this.reservedSectorCount - 1) * this.sectorSize),
    # file allocation tables
    Array(this.fatCount, Array(this.sectorsPerFat * this.sectorSize / FatEntry.sizeof(), FatEntry)),
    # root directories
    Array((this.rootdirEntryCount * DirEntry.sizeof()) / this.sectorSize, DirEntry),
)


class File(object):
    def __init__(self, dirEntry, fs):
        self.fs = fs
        self.dirEntry = dirEntry

    @classmethod
    def fromDirEntry(cls, dirEntry, fs):
        if dirEntry.name[0] in "\x00\xe5\x2e":
            return None
        a = dirEntry.attributes
        #Long file name directory entry
        if a.volumeLabel and a.system and a.hidden and a.readonly:
            return None
        if a.subDirectory:
            return Directory(dirEntry, fs)
        return File(dirEntry, fs)

    @classmethod
    def fromDirEntries(cls, dirEntries, fs):
        return filter(None, [cls.fromDirEntry(de, fs) for de in dirEntries])

    def toStream(self, stream):
        self.fs.fileToStream(self.dirEntry.firstCluster, stream)
    
    @property
    def name(self):
        return "%s.%s" % (self.dirEntry.name.rstrip(), self.dirEntry.extension)
    
    def __str__(self):
        return "&%s %s" % (self.dirEntry.firstCluster, self.name)


class Directory(File):
    def __init__(self, dirEntry, fs, children=None):
        super(Directory, self).__init__(dirEntry, fs)
        self.children = children
        if not self.children:
            self.children = File.fromDirEntries(\
                self.fs.getDirEntries(\
                self.dirEntry.firstCluster), fs)
    
    @property
    def name(self):
        return self.dirEntry.name.rstrip()

    def __str__(self):
        return "&%s %s/" % (self.dirEntry.firstCluster, self.name)

    def __getitem__(self, name):
        for file in self.children:
            if file.name == name:
                return file
    
    def __iter__(self):
        return iter(self.children)


class FatFs(Directory):
    def __init__(self, stream):
        self.stream = stream
        self.pdr = PreDataRegion.parse_stream(stream)
        super(FatFs, self).__init__(dirEntry = None, 
                fs = self, children = File.fromDirEntries(
                        self.pdr.rootdirs, self))

    def fileToStream(self, clidx, stream):
        for clidx in self.getLinkedClusters(clidx):
            self.clusterToStream(clidx, stream)

    def clusterToStream(self, clidx, stream):
        start, todo = self.getClusterSlice(clidx)
        self.stream.seek(start)
        while todo > 0:
            read = self.stream.read(todo)
            if not len(read):
                print("failed to read %s bytes at %s" % (todo, self.stream.tell()))
                raise EOFError()
            todo -= len(read)
            stream.write(read)

    def getClusterSlice(self, clidx):
        startSector = self.pdr.reservedSectorCount \
                + self.pdr.fatCount * self.pdr.sectorsPerFat \
                + (self.pdr.rootdirEntryCount * 32) \
                    / self.pdr.sectorSize \
                + (clidx-2) * self.pdr.sectorsPerCluster
        start = startSector * self.pdr.sectorSize
        length = self.pdr.sectorSize * self.pdr.sectorsPerCluster
        return (start, length)

    def getLinkedClusters(self, clidx):
        res = []
        while clidx != "last_cluster":
            if not isinstance(clidx, numbers.Real):
                print(clidx)
                assert False
            assert 2 <= clidx <= 0xffef
            res.append(clidx)
            clidx = self.getNextCluster(clidx)
            assert clidx not in res
        return res

    def getNextCluster(self, clidx):
        ress = set([fat[clidx] for fat in self.pdr.fats])
        if len(ress)==1:
            return ress.pop()
        print("inconsistencie between FATs:  %s points to" % (clidx,))
        for i,fat in enumerate(self.pdr.fats):
            print("\t%s according to fat #%s" % (fat[clidx], i))
        res = ress.pop()
        print ("assuming %s" % res)
        return res
    
    def getDirEntries(self, clidx):
        try:
            for de in self._getDirEntries(clidx):
                yield de
        except IOError:
            print("failed to read directory entries at %s" % clidx)

    def _getDirEntries(self, clidx):
        de = DirEntry("dirEntry")
        with BytesIO() as mem:
            self.fileToStream(clidx, mem)
            mem.seek(0)
            with BufferedReader(mem) as reader:
                while reader.peek(1):
                    yield de.parse_stream(reader)
    def __str__(self):
        return "/"

    @property
    def name(self):
        return ""



from declarativeunittest import *
from construct import *
from construct.lib import *


def test_multiprocessing():
    import multiprocessing

    def worker(q):
        obj = q.get()
        print(obj)

    queue = multiprocessing.Queue()

    p = multiprocessing.Process(target=worker, args=(queue,))
    p.start()

    obj = Container(name="test")
    print(obj)
    queue.put(obj)

    # Wait for the worker to finish
    queue.close()
    queue.join_thread()
    p.join()



from construct import *

docs = """
PE/COFF format as used on Windows to store EXE DLL SYS files. This includes 64-bit and .NET code.

Microsoft specifications:
https://msdn.microsoft.com/en-us/library/windows/desktop/ms680547(v=vs.85).aspx
https://msdn.microsoft.com/en-us/library/ms809762.aspx
Format tutorial breakdown at:
http://blog.dkbza.org/
https://drive.google.com/file/d/0B3_wGJkuWLytQmc2di0wajB1Xzg/view
https://drive.google.com/file/d/0B3_wGJkuWLytbnIxY1J5WUs4MEk/view

Authored by Arkadiusz Bulski, under same license.
"""

msdosheader = Struct(
    "signature" / Const(b"MZ"),
    "lfanew" / Pointer(0x3c, Int16ul),
)

coffheader = Struct(
    "signature" / Const(b"PE\x00\x00"),
    "machine" / Enum(Int16ul,
        UNKNOWN = 0x0,
        AM33 = 0x1d3,
        AMD64 = 0x8664,
        ARM = 0x1c0,
        ARM64 = 0xaa64,
        ARMNT = 0x1c4,
        EBC = 0xebc,
        I386 = 0x14c,
        IA64 = 0x200,
        M32R = 0x9041,
        MIPS16 = 0x266,
        MIPSFPU = 0x366,
        MIPSFPU16 = 0x466,
        POWERPC = 0x1f0,
        POWERPCFP = 0x1f1,
        R4000 = 0x166,
        RISCV32 = 0x5032,
        RISCV64 = 0x5064,
        RISCV128 = 0x5128,
        SH3 = 0x1a2,
        SH3DSP = 0x1a3,
        SH4 = 0x1a6,
        SH5 = 0x1a8,
        THUMB = 0x1c2,
        WCEMIPSV2 = 0x169,
    ),
    "sections_count" / Int16ul,
    "created" / Timestamp(Int32ul, 1., 1970),
    "symbol_pointer" / Int32ul, #deprecated
    "symbol_count" / Int32ul, #deprecated
    "optionalheader_size" / Int16ul,
    "characteristics" / FlagsEnum(Int16ul,
        RELOCS_STRIPPED = 0x0001,
        EXECUTABLE_IMAGE = 0x0002,
        LINE_NUMS_STRIPPED = 0x0004, #deprecated
        LOCAL_SYMS_STRIPPED = 0x0008, #deprecated
        AGGRESSIVE_WS_TRIM = 0x0010, #deprecated
        LARGE_ADDRESS_AWARE = 0x0020,
        RESERVED = 0x0040, #reserved
        BYTES_REVERSED_LO = 0x0080, #deprecated
        MACHINE_32BIT = 0x0100,
        DEBUG_STRIPPED = 0x0200,
        REMOVABLE_RUN_FROM_SWAP = 0x0400,
        SYSTEM = 0x1000,
        DLL = 0x2000,
        UNIPROCESSOR_ONLY = 0x4000,
        BIG_ENDIAN_MACHINE = 0x8000, #deprecated
    ),
)

plusfield = IfThenElse(this.signature == "PE32plus", Int64ul, Int32ul)

entriesnames = {
    0 : 'export_table',
    1 : 'import_table',
    2 : 'resource_table',
    3 : 'exception_table',
    4 : 'certificate_table',
    5 : 'base_relocation_table',
    6 : 'debug',
    7 : 'architecture',
    8 : 'global_ptr',
    9 : 'tls_table',
    10 : 'load_config_table',
    11 : 'bound_import',
    12 : 'import_address_table',
    13 : 'delay_import_descriptor',
    14 : 'clr_runtime_header',
    15 : 'reserved',
}

datadirectory = Struct(
    "name" / Computed(lambda this: entriesnames[this._._index]),
    "virtualaddress" / Int32ul,
    "size" / Int32ul,
)

optionalheader = Struct(
    "signature" / Enum(Int16ul, 
        PE32 = 0x10b,
        PE32plus = 0x20b,
        ROMIMAGE = 0x107,
    ),
    "linker_version" / Int8ul[2],
    "size_code" / Int32ul,
    "size_initialized_data" / Int32ul,
    "size_uninitialized_data" / Int32ul,
    "entrypoint" / Int32ul,
    "base_code" / Int32ul,
    "base_data" / If(this.signature == "PE32", Int32ul),
    "image_base" / plusfield,
    "section_alignment" / Int32ul,
    "file_alignment" / Int32ul,
    "os_version" / Int16ul[2],
    "image_version" / Int16ul[2],
    "subsystem_version" / Int16ul[2],
    "win32versionvalue" / Int32ul, #deprecated
    "image_size" / Int32ul,
    "headers_size" / Int32ul,
    "checksum" / Int32ul,
    "subsystem" / Enum(Int16ul,
        UNKNOWN = 0,
        NATIVE = 1,
        WINDOWS_GUI = 2,
        WINDOWS_CUI = 3,
        OS2_CUI = 5,
        POSIX_CUI = 7,
        WINDOWS_NATIVE = 8,
        WINDOWS_CE_GUI = 9,
        EFI_APPLICATION = 10,
        EFI_BOOT_SERVICE_DRIVER = 11,
        EFI_RUNTIME_DRIVER = 12,
        EFI_ROM = 13,
        XBOX = 14,
        WINDOWS_BOOT_APPLICATION = 16,
    ),
    "dll_characteristics" / FlagsEnum(Int16ul,
        HIGH_ENTROPY_VA = 0x0020,
        DYNAMIC_BASE = 0x0040,
        FORCE_INTEGRITY = 0x0080,
        NX_COMPAT = 0x0100,
        NO_ISOLATION = 0x0200,
        NO_SEH = 0x0400,
        NO_BIND = 0x0800,
        APPCONTAINER = 0x1000,
        WDM_DRIVER = 0x2000,
        GUARD_CF = 0x4000,
        TERMINAL_SERVER_AWARE = 0x8000,
    ),
    "stack_reserve" / plusfield,
    "stack_commit" / plusfield,
    "heap_reserve" / plusfield,
    "heap_commit" / plusfield,
    "loader_flags" / Int32ul, #reserved
    "datadirectories_count" / Int32ul,
    "datadirectories" / Array(this.datadirectories_count, 
        datadirectory),
)

section = Struct(
    "name" / PaddedString(8, "utf8"),
    "virtual_size" / Int32ul,
    "virtual_address" / Int32ul,
    "rawdata_size" / Int32ul,
    "rawdata_pointer" / Int32ul,
    "relocations_pointer" / Int32ul,
    "linenumbers_pointer" / Int32ul,
    "relocations_count" / Int16ul,
    "linenumbers_count" / Int16ul,
    "characteristics" / FlagsEnum(Int32ul,
        TYPE_REG = 0x00000000,
        TYPE_DSECT = 0x00000001,
        TYPE_NOLOAD = 0x00000002,
        TYPE_GROUP = 0x00000004,
        TYPE_NO_PAD = 0x00000008,
        TYPE_COPY = 0x00000010,
        CNT_CODE = 0x00000020,
        CNT_INITIALIZED_DATA = 0x00000040,
        CNT_UNINITIALIZED_DATA = 0x00000080,
        LNK_OTHER = 0x00000100,
        LNK_INFO = 0x00000200,
        TYPE_OVER = 0x00000400,
        LNK_REMOVE = 0x00000800,
        LNK_COMDAT = 0x00001000,
        MEM_FARDATA = 0x00008000,
        MEM_PURGEABLE = 0x00020000,
        MEM_16BIT = 0x00020000,
        MEM_LOCKED = 0x00040000,
        MEM_PRELOAD = 0x00080000,
        ALIGN_1BYTES = 0x00100000,
        ALIGN_2BYTES = 0x00200000,
        ALIGN_4BYTES = 0x00300000,
        ALIGN_8BYTES = 0x00400000,
        ALIGN_16BYTES = 0x00500000,
        ALIGN_32BYTES = 0x00600000,
        ALIGN_64BYTES = 0x00700000,
        ALIGN_128BYTES = 0x00800000,
        ALIGN_256BYTES = 0x00900000,
        ALIGN_512BYTES = 0x00A00000,
        ALIGN_1024BYTES = 0x00B00000,
        ALIGN_2048BYTES = 0x00C00000,
        ALIGN_4096BYTES = 0x00D00000,
        ALIGN_8192BYTES = 0x00E00000,
        LNK_NRELOC_OVFL = 0x01000000,
        MEM_DISCARDABLE = 0x02000000,
        MEM_NOT_CACHED = 0x04000000,
        MEM_NOT_PAGED = 0x08000000,
        MEM_SHARED = 0x10000000,
        MEM_EXECUTE = 0x20000000,
        MEM_READ = 0x40000000,
        MEM_WRITE = 0x80000000,
    ),
    "rawdata" / Pointer(this.rawdata_pointer, 
        Bytes(lambda this: this.rawdata_size if this.rawdata_pointer else 0)),
    "relocations" / Pointer(this.relocations_pointer,
        Array(this.relocations_count, Struct(
            "virtualaddress" / Int32ul,
            "symboltable_index" / Int32ul,
            "type" / Int16ul * "complicated platform-dependant Enum",
        ))
    ),
    "linenumbers" / Pointer(this.linenumbers_pointer,
        Array(this.linenumbers_count, Struct(
            "_type" / Int32ul,
            "_linenumber" / Int16ul,
            "is_symboltableindex" / Computed(this._linenumber == 0),
            "is_linenumber" / Computed(this._linenumber > 0),
            "symboltableindex" / If(this.is_symboltableindex, Computed(this._type)),
            "linenumber" / If(this.is_linenumber, Computed(this._linenumber)),
            "virtualaddress" / If(this.is_linenumber, Computed(this._type)),
        ))
    ),
)

pe32file = docs * Struct(
    "msdosheader" / msdosheader,
    Seek(this.msdosheader.lfanew),
    "coffheader" / coffheader,
    "optionalheader" / If(this.coffheader.optionalheader_size > 0, optionalheader),
    "sections_count" / Computed(this.coffheader.sections_count),
    "sections" / Array(this.sections_count, section),
)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
from kaitaistruct import __version__ as ks_version, KaitaiStruct, KaitaiStream, BytesIO


if parse_version(ks_version) < parse_version('0.7'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.7 or later is required, but you have %s" % (ks_version))

class Comparison1Kaitai(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.count = self._io.read_u4le()
        self.items = [None] * (self.count)
        for i in range(self.count):
            self.items[i] = self._root.Item(self._io, self, self._root)


    class Item(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.num1 = self._io.read_u1()
            self.num2_lo = self._io.read_u2le()
            self.num2_hi = self._io.read_u1()
            # tweaked
            self.num2
            self.flags = self._root.Item.Flags(self._io, self, self._root)
            self.fixedarray1 = [None] * (3)
            for i in range(3):
                self.fixedarray1[i] = self._io.read_u1()

            self.name1 = (self._io.read_bytes_term(0, False, True, True)).decode(u"utf-8")
            self.len_name2 = self._io.read_u1()
            self.name2 = (self._io.read_bytes(self.len_name2)).decode(u"utf-8")

        class Flags(KaitaiStruct):
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._read()

            def _read(self):
                self.bool1 = self._io.read_bits_int(1) != 0
                self.num4 = self._io.read_bits_int(3)
                self.padding = self._io.read_bits_int(4)


        @property
        def num2(self):
            if hasattr(self, '_m_num2'):
                return self._m_num2 if hasattr(self, '_m_num2') else None

            self._m_num2 = ((self.num2_hi << 16) | self.num2_lo)
            return self._m_num2 if hasattr(self, '_m_num2') else None


data = open("blob","rb").read()
Comparison1Kaitai.from_bytes(data)

from timeit import timeit
parsetime = timeit(lambda: Comparison1Kaitai.from_bytes(data), number=1000)/1000
print("Timeit measurements:")
print("parsing:           {:.10f} sec/call".format(parsetime))



from io import BlockingIOError
from time import sleep
from sys import maxsize


class RestreamedBytesIO(object):

    def __init__(self, substream, decoder, decoderunit, encoder, encoderunit):
        self.substream = substream
        self.encoder = encoder
        self.encoderunit = encoderunit
        self.decoder = decoder
        self.decoderunit = decoderunit
        self.rbuffer = b""
        self.wbuffer = b""
        self.sincereadwritten = 0

    def read(self, count=None):
        if count is None:
            while True:
                data = self.substream.read(self.decoderunit)
                if data is None or len(data) == 0:
                    break
                self.rbuffer += self.decoder(data)
            data, self.rbuffer = self.rbuffer, b''
            self.sincereadwritten += len(data)
            return data

        else:
            if count < 0:
                raise ValueError("count cannot be negative")
            while len(self.rbuffer) < count:
                data = self.substream.read(self.decoderunit)
                if data is None or len(data) == 0:
                    return b''
                self.rbuffer += self.decoder(data)
            data, self.rbuffer = self.rbuffer[:count], self.rbuffer[count:]
            self.sincereadwritten += count
            return data

    def write(self, data):
        self.wbuffer += data
        datalen = len(data)
        while len(self.wbuffer) >= self.encoderunit:
            data, self.wbuffer = self.wbuffer[:self.encoderunit], self.wbuffer[self.encoderunit:]
            self.substream.write(self.encoder(data))
        self.sincereadwritten += datalen
        return datalen

    def close(self):
        if len(self.rbuffer):
            raise ValueError("closing stream but %d unread bytes remain, %d is decoded unit" % (len(self.rbuffer), self.decoderunit))
        if len(self.wbuffer):
            raise ValueError("closing stream but %d unwritten bytes remain, %d is encoded unit" % (len(self.wbuffer), self.encoderunit))

    def seek(self, at, whence=0):
        if whence == 0 and at == self.sincereadwritten:
            pass
        else:
            raise IOError

    def seekable(self):
        return False

    def tell(self):
        """WARNING: tell is correct only on read-only and write-only instances."""
        return self.sincereadwritten

    def tellable(self):
        return True


class RebufferedBytesIO(object):

    def __init__(self, substream, tailcutoff=None):
        self.substream = substream
        self.offset = 0
        self.rwbuffer = b""
        self.moved = 0
        self.tailcutoff = tailcutoff

    def read(self, count=None):
        if count is None:
            raise ValueError("count must be integer, reading until EOF not supported")
        startsat = self.offset
        endsat = startsat + count
        if startsat < self.moved:
            raise IOError("could not read because tail was cut off")
        while self.moved + len(self.rwbuffer) < endsat:
            try:
                newdata = self.substream.read(128*1024)
            except BlockingIOError:
                newdata = None
            if not newdata:
                sleep(0)
                continue
            self.rwbuffer += newdata
        data = self.rwbuffer[startsat-self.moved:endsat-self.moved]
        self.offset += count
        if self.tailcutoff is not None and self.moved < self.offset - self.tailcutoff:
            removed = self.offset - self.tailcutoff - self.moved
            self.moved += removed
            self.rwbuffer = self.rwbuffer[removed:]
        if len(data) < count:
            raise IOError("could not read enough bytes, something went wrong")
        return data

    def write(self, data):
        startsat = self.offset
        endsat = startsat + len(data)
        while self.moved + len(self.rwbuffer) < startsat:
            newdata = self.substream.read(128*1024)
            self.rwbuffer += newdata
            if not newdata:
                sleep(0)
        self.rwbuffer = self.rwbuffer[:startsat-self.moved] + data + self.rwbuffer[endsat-self.moved:]
        self.offset = endsat
        if self.tailcutoff is not None and self.moved < self.offset - self.tailcutoff:
            removed = self.offset - self.tailcutoff - self.moved
            self.moved += removed
            self.rwbuffer = self.rwbuffer[removed:]
        return len(data)

    def seek(self, at, whence=0):
        if whence == 0:
            self.offset = at
            return self.offset
        elif whence == 1:
            self.offset += at
            return self.offset
        else:
            raise ValueError("this class seeks only with whence: 0 and 1 (excluded 2)")

    def seekable(self):
        return True

    def tell(self):
        return self.offset

    def tellable(self):
        return True

    def cachedfrom(self):
        return self.moved

    def cachedto(self):
        return self.moved + len(self.rwbuffer)



from declarativeunittest import *
from construct import *
from construct.lib import *

from deprecated_gallery import  *


def test_png():
    commondumpdeprecated(png_file, "sample.png")

def test_emf():
    commondumpdeprecated(emf_file, "emf1.emf")

def test_bitmap():
    commondumpdeprecated(bitmap_file, "bitmap1.bmp")
    commondumpdeprecated(bitmap_file, "bitmap4.bmp")
    commondumpdeprecated(bitmap_file, "bitmap8.bmp")
    commondumpdeprecated(bitmap_file, "bitmap24.bmp")

def test_wmf():
    commondumpdeprecated(wmf_file, "wmf1.wmf")

def test_gif():
    commondumpdeprecated(gif_file, "sample.gif")

def test_mbr():
    commondumpdeprecated(mbr_format, "mbr1")

def test_cap():
    commondumpdeprecated(cap_file, "cap2.cap")

def test_snoop():
    commondumpdeprecated(snoop_file, "snoop1")

def test_pe32():
    commondumpdeprecated(pe32_file, "python.exe")
    commondumpdeprecated(pe32_file, "NOTEPAD.EXE")
    commondumpdeprecated(pe32_file, "sqlite3.dll")

def test_elf32():
    commondumpdeprecated(elf32_file, "ctypes.so")



"""
Executable and Linkable Format (ELF), 32 bit, big or little endian.
Used on Unix systems as a replacement of the older a.out format.

Big-endian support kindly submitted by Craig McQueen (mcqueen-c#edsrd1!yzk!co!jp).
"""

from construct import *


def elf32_body(ElfInt16, ElfInt32):

    elf32_program_header = Struct(
        "type" / Enum(ElfInt32,
            NULL = 0,
            LOAD = 1,
            DYNAMIC = 2,
            INTERP = 3,
            NOTE = 4,
            SHLIB = 5,
            PHDR = 6,
        ),
        "offset" / ElfInt32,
        "vaddr" / ElfInt32,
        "paddr" / ElfInt32,
        "file_size" / ElfInt32,
        "mem_size" / ElfInt32,
        "flags" / ElfInt32,
        "align" / ElfInt32,
    )
    
    elf32_section_header = Struct(
        "name_offset" / ElfInt32,
        "name" / Pointer(this._.strtab_data_offset + this.name_offset, 
            CString("utf8")),
        "type" / Enum(ElfInt32, 
            NULL = 0,
            PROGBITS = 1,
            SYMTAB = 2,
            STRTAB = 3,
            RELA = 4,
            HASH = 5,
            DYNAMIC = 6,
            NOTE = 7,
            NOBITS = 8,
            REL = 9,
            SHLIB = 10,
            DYNSYM = 11,
        ),
        "flags" / ElfInt32,
        "addr" / ElfInt32,
        "offset" / ElfInt32,
        "size" / ElfInt32,
        "link" / ElfInt32,
        "info" / ElfInt32,
        "align" / ElfInt32,
        "entry_size" / ElfInt32,
        "data" / Pointer(this.offset, 
            Bytes(this.size)),
    )
    
    return Struct(
        "type" / Enum(ElfInt16,
            NONE = 0,
            RELOCATABLE = 1,
            EXECUTABLE = 2,
            SHARED = 3,
            CORE = 4,
        ),
        "machine" / Enum(ElfInt16,
            NONE = 0,
            M32 = 1,
            SPARC = 2,
            I386 = 3,
            Motorolla68K = 4,
            Motorolla88K = 5,
            Intel860 = 7,
            MIPS = 8,
        ),
        "version"              / ElfInt32,
        "entry"                / ElfInt32,
        "ph_offset"            / ElfInt32,
        "sh_offset"            / ElfInt32,
        "flags"                / ElfInt32,
        "header_size"          / ElfInt16,
        "ph_entry_size"        / ElfInt16,
        "ph_count"             / ElfInt16,
        "sh_entry_size"        / ElfInt16,
        "sh_count"             / ElfInt16,
        "strtab_section_index" / ElfInt16,
        
        # calculate the string table data offset (pointer arithmetics)
        # ugh... anyway, we need it in order to read the section names, later on
        "strtab_data_offset" / Pointer(this.sh_offset + this.strtab_section_index * this.sh_entry_size + 16, 
            ElfInt32),
        
        "program_table" / Pointer(this.ph_offset, 
            elf32_program_header[this.ph_count]),
        
        "sections" / Pointer(this.sh_offset, 
            elf32_section_header[this.sh_count]),
    )

elf32_file = Struct(
    "identifier" / Struct(
        Const(b"\x7fELF"),
        "file_class" / Enum(Byte,
            NONE = 0,
            CLASS32 = 1,
            CLASS64 = 2,
        ),
        "encoding" / Enum(Byte,
            NONE = 0,
            LSB = 1,
            MSB = 2,            
        ),
        "version" / Byte,
        Padding(9),
    ),
    "body" / IfThenElse(this.identifier.encoding == "LSB",
        elf32_body(Int16ul, Int32ul),
        elf32_body(Int16ub, Int32ub),
    ),
)



"""
Contributed by Dany Zatuchna (danzat at gmail)

Implementation of the following grammar for the GIF89a file format

<GIF Data Stream> ::=     Header <Logical Screen> <Data>* Trailer
<Logical Screen> ::=      Logical Screen Descriptor [Global Color Table]
<Data> ::=                <Graphic Block>  | <Special-Purpose Block>
<Graphic Block> ::=       [Graphic Control Extension] <Graphic-Rendering Block>
<Graphic-Rendering Block> ::=  <Table-Based Image>  | Plain Text Extension
<Table-Based Image> ::=   Image Descriptor [Local Color Table] Image Data
<Special-Purpose Block> ::=    Application Extension  | Comment Extension
"""

from construct import *


data_sub_block = PascalString(Int8ul, "utf8")

gif_logical_screen = Struct(
    "width" / Int16ul,
    "height" / Int16ul,
    "flags" / BitStruct(
        "global_color_table" / Bit,
        "color_resolution" / BitsInteger(3),
        "sort_flag" / Bit,
        "global_color_table_bpp" / BitsInteger(3),
    ),
    "bgcolor_index" / Int8ul,
    "pixel_aspect_ratio" / Int8ul,
    "palette" / If(this.flags.global_color_table,
        Array(lambda this: 2**(this.flags.global_color_table_bpp + 1),
            Struct(
                "R" / Int8ul,
                "G" / Int8ul,
                "B" / Int8ul,
            ))),
)

application_extension = Struct(
    "block_size" / Const(11, Int8ul),
    "application_identifier" / PaddedString(8, "utf8"),
    "application_auth_code" / PaddedString(3, "utf8"),
    "data_sub_block" / data_sub_block,
    "block_terminator" / Int8ul,
)

comment_extension = Struct(
    "data_sub_block" / data_sub_block,
    "block_terminator" / Int8ul,
)

graphic_control_extension = Struct(
    "block_size" / Const(4, Int8ul),
    "flags" / BitStruct(
        "reserved" / BitsInteger(3),
        "disposal_method" / BitsInteger(3),
        "user_input_flag" / Bit,
        "transparent_color_flag" / Bit,
    ),
    "delay" / Int16ul,
    "transparent_color_index" / Int8ul,
    "block_terminator" / Int8ul,
)

plain_text_extension = Struct(
    "block_size" / Const(12, Int8ul),
    "text_left" / Int16ul,
    "text_top" / Int16ul,
    "text_width" / Int16ul,
    "text_height" / Int16ul,
    "cell_width" / Int8ul,
    "cell_height" / Int8ul,
    "foreground_index" / Int8ul,
    "background_index" / Int8ul,
    "data_sub_block" / data_sub_block,
    "block_terminator" / Int8ul,
)

extension = Struct(
    "label" / Int8ul,
    "ext" / Switch(this.label, {
        0xFF: application_extension,
        0xFE: comment_extension,
        0xF9: graphic_control_extension,
        0x01: plain_text_extension
    }),
)

image_descriptor = Struct(
    "left" / Int16ul,
    "top" / Int16ul,
    "width" / Int16ul,
    "height" / Int16ul,
    "flags" / BitStruct(
        "local_color_table" / Bit,
        "interlace" / Bit,
        "sort" / Bit,
        "reserved" / BitsInteger(2),
        "local_color_table_bpp" / BitsInteger(3),
    ),
    "palette" / If(this.flags.local_color_table,
        Array(lambda this: 2**(this.flags.local_color_table_bpp + 1),
            Struct(
                "R" / Int8ul,
                "G" / Int8ul,
                "B" / Int8ul,
            ))),
    "lzw_minimum_code_size" / Int8ul,
    "data_sub_block" / RepeatUntil(lambda obj,lst,ctx: obj.size == 0, data_sub_block),
)

gif_data = Struct(
    "introducer" / Int8ul,
    "data" / Switch(this.introducer, {
        0x21: extension,
        0x2C: image_descriptor
    }),
)

gif_file = Struct(
    "signature" / Const(b"GIF"),
    "version" / Const(b"89a"),
    "logical_screen" / gif_logical_screen,
    "data" / GreedyRange(gif_data),
    # Const(Int8ul("trailer"), 0x3B)
)



from declarativeunittest import *
from construct import *
from construct.lib import *


def test_getitem():
    c = Container(a=1)
    assert c["a"] == 1
    assert c.a == 1
    assert raises(lambda: c.unknownkey) == AttributeError
    assert raises(lambda: c["unknownkey"]) == KeyError

def test_setitem():
    c = Container()
    c.a = 1
    assert c["a"] == 1
    assert c.a == 1
    c["a"] = 2
    assert c["a"] == 2
    assert c.a == 2

def test_delitem():
    c = Container(a=1, b=2)
    del c.a
    assert "a" not in c
    assert raises(lambda: c.a) == AttributeError
    assert raises(lambda: c["a"]) == KeyError
    del c["b"]
    assert "b" not in c
    assert raises(lambda: c.b) == AttributeError
    assert raises(lambda: c["b"]) == KeyError
    assert c == Container()
    assert list(c) == []

def test_ctor_empty():
    c = Container()
    assert len(c) == 0
    assert list(c.items()) == []
    assert c == Container()
    assert c == Container(c)
    assert c == Container({})
    assert c == Container([])

def test_ctor_chained():
    c = Container(a=1)(b=2)(c=3)(d=4)
    assert c == Container(c)

def test_ctor_dict():
    c = Container(a=1)(b=2)(c=3)(d=4)
    c = Container(c)
    assert len(c) == 4
    assert list(c.items()) == [('a',1),('b',2),('c',3),('d',4)]

def test_ctor_seqoftuples():
    c = Container([('a',1),('b',2),('c',3),('d',4)])
    assert len(c) == 4
    assert list(c.items()) == [('a',1),('b',2),('c',3),('d',4)]

def test_ctor_orderedkw():
    c = Container(a=1, b=2, c=3, d=4)
    d = Container(a=1)(b=2)(c=3)(d=4)
    assert c == d
    assert len(c) == len(d)
    assert list(c.items()) == list(d.items())

def test_keys():
    c = Container(a=1)(b=2)(c=3)(d=4)
    assert list(c.keys()) == ["a","b","c","d"]

def test_values():
    c = Container(a=1)(b=2)(c=3)(d=4)
    assert list(c.values()) == [1,2,3,4]

def test_items():
    c = Container(a=1)(b=2)(c=3)(d=4)
    assert list(c.items()) == [("a",1),("b",2),("c",3),("d",4)]

def test_iter():
    c = Container(a=1)(b=2)(c=3)(d=4)
    assert list(c) == list(c.keys())

def test_clear():
    c = Container(a=1)(b=2)(c=3)(d=4)
    c.clear()
    assert c == Container()
    assert list(c.items()) == []

def test_pop():
    c = Container(a=1)(b=2)(c=3)(d=4)
    assert c.pop("b") == 2
    assert c.pop("d") == 4
    assert c.pop("a") == 1
    assert c.pop("c") == 3
    assert raises(c.pop, "missing") == KeyError
    assert c == Container()

def test_popitem():
    c = Container(a=1)(b=2)(c=3)(d=4)
    assert c.popitem() == ("d",4)
    assert c.popitem() == ("c",3)
    assert c.popitem() == ("b",2)
    assert c.popitem() == ("a",1)
    assert raises(c.popitem) == KeyError

def test_update_dict():
    c = Container(a=1)(b=2)(c=3)(d=4)
    d = Container()
    d.update(c)
    assert d.a == 1
    assert d.b == 2
    assert d.c == 3
    assert d.d == 4
    assert c == d
    assert list(c.items()) == list(d.items())

def test_update_seqoftuples():
    c = Container(a=1)(b=2)(c=3)(d=4)
    d = Container()
    d.update([("a",1),("b",2),("c",3),("d",4)])
    assert d.a == 1
    assert d.b == 2
    assert d.c == 3
    assert d.d == 4
    assert c == d
    assert list(c.items()) == list(d.items())

def test_copy_method():
    c = Container(a=1)
    d = c.copy()
    assert c == d
    assert c is not d

def test_copy():
    from copy import copy, deepcopy

    c = Container(a=1)
    d = copy(c)
    assert c == d
    assert c is not d

def test_deepcopy():
    from copy import copy, deepcopy

    c = Container(a=1)
    d = deepcopy(c)
    d.a = 2
    assert c != d
    assert c is not d

def test_pickling():
    import pickle

    empty = Container()
    empty_unpickled = pickle.loads(pickle.dumps(empty))
    assert empty_unpickled == empty

    nested = Container(a=1)(b=Container())(c=3)(d=Container(e=4))
    nested_unpickled = pickle.loads(pickle.dumps(nested))
    assert nested_unpickled == nested

def test_eq_issue_818():
    c = Container(a=1)(b=2)(c=3)(d=4)(e=5)
    d = Container(a=1)(b=2)(c=3)(d=4)(e=5)
    assert c == c
    assert d == d
    assert c == d
    assert d == c

    a = Container(a=1)(b=2)
    b = Container(a=1)(b=2)(c=3)
    assert not a == b
    assert not b == a

    # c contains internal '_io' field, which shouldn't be considered in the comparison
    c = Struct('a' / Int8ul).parse(b'\x01')
    d = {'a': 1}
    assert c == d
    assert d == c

def test_eq_numpy():
    import numpy
    c = Container(arr=numpy.zeros(10, dtype=numpy.uint8))
    d = Container(arr=numpy.zeros(10, dtype=numpy.uint8))
    assert c == d

def test_ne_issue_818():
    c = Container(a=1)(b=2)(c=3)
    d = Container(a=1)(b=2)(c=3)(d=4)(e=5)
    assert c != d
    assert d != c

def test_str_repr_empty():
    c = Container()
    assert str(c) == "Container: "
    assert repr(c) == "Container()"
    assert eval(repr(c)) == c

def test_str_repr():
    c = Container(a=1)(b=2)(c=3)
    assert str(c) == "Container: \n    a = 1\n    b = 2\n    c = 3"
    assert repr(c) == "Container(a=1, b=2, c=3)"
    assert eval(repr(c)) == c

def test_str_repr_nested():
    c = Container(a=1)(b=2)(c=Container())
    assert str(c) == "Container: \n    a = 1\n    b = 2\n    c = Container: "
    assert repr(c) == "Container(a=1, b=2, c=Container())"
    assert eval(repr(c)) == c

def test_str_repr_recursive():
    c = Container(a=1)(b=2)
    c.c = c
    assert str(c) == "Container: \n    a = 1\n    b = 2\n    c = <recursion detected>"
    assert repr(c) == "Container(a=1, b=2, c=<recursion detected>)"

def test_fullstrings():
    setGlobalPrintFullStrings(True)
    c = Container(data=b"1234567890")
    assert str(c) == "Container: \n    data = b'1234567890' (total 10)"
    assert repr(c) == "Container(data=b'1234567890')"
    c = Container(data=u"1234567890")
    assert str(c) == "Container: \n    data = u'1234567890' (total 10)"
    assert repr(c) == "Container(data=u'1234567890')"
    c = Container(data=b"1234567890123456789012345678901234567890")
    assert str(c) == "Container: \n    data = b'1234567890123456789012345678901234567890' (total 40)"
    assert repr(c) == "Container(data=b'1234567890123456789012345678901234567890')"
    c = Container(data=u"1234567890123456789012345678901234567890")
    assert str(c) == "Container: \n    data = u'1234567890123456789012345678901234567890' (total 40)"
    assert repr(c) == "Container(data=u'1234567890123456789012345678901234567890')"

    setGlobalPrintFullStrings(False)
    c = Container(data=b"1234567890")
    assert str(c) == "Container: \n    data = b'1234567890' (total 10)"
    assert repr(c) == "Container(data=b'1234567890')"
    c = Container(data=u"1234567890")
    assert str(c) == "Container: \n    data = u'1234567890' (total 10)"
    assert repr(c) == "Container(data=u'1234567890')"
    c = Container(data=b"1234567890123456789012345678901234567890")
    assert str(c) == "Container: \n    data = b'1234567890123456'... (truncated, total 40)"
    assert repr(c) == "Container(data=b'1234567890123456789012345678901234567890')"
    c = Container(data=u"1234567890123456789012345678901234567890")
    assert str(c) == "Container: \n    data = u'12345678901234567890123456789012'... (truncated, total 40)"
    assert repr(c) == "Container(data=u'1234567890123456789012345678901234567890')"

    setGlobalPrintFullStrings()

def test_falseflags():
    d = FlagsEnum(Byte, set=1, unset=2)
    c = d.parse(b"\x01")

    setGlobalPrintFalseFlags(True)
    assert str(c) == "Container: \n    set = True\n    unset = False"
    assert repr(c) == "Container(set=True, unset=False)"

    setGlobalPrintFalseFlags(False)
    assert str(c) == "Container: \n    set = True"
    assert repr(c) == "Container(set=True, unset=False)"

    setGlobalPrintFalseFlags()

def test_privateentries():
    c = Container(_private = 1)

    setGlobalPrintPrivateEntries(True)
    assert str(c) == "Container: \n    _private = 1"
    assert repr(c) == "Container()"

    setGlobalPrintPrivateEntries(False)
    assert str(c) == "Container: "
    assert repr(c) == "Container()"

    setGlobalPrintPrivateEntries()

def test_len_bool():
    c = Container(a=1)(b=2)(c=3)(d=4)
    assert len(c) == 4
    assert c
    c = Container()
    assert len(c) == 0
    assert not c

def test_in():
    c = Container(a=1)
    assert "a" in c
    assert "b" not in c

def test_regression_recursionlock():
    print("REGRESSION: recursion_lock() used to leave private keys.")
    c = Container()
    str(c); repr(c)
    assert not c



##############################################################
# WARNING: HEADER IS SKIPPED NOT PARSED, DATETIME CAN BE WRONG
# 
# https://wiki.wireshark.org/Development/LibpcapFileFormat
##############################################################

from construct import *
import time, datetime


# use core Timestamp
class MicrosecAdapter(Adapter):
    def _decode(self, obj, context, path):
        return datetime.datetime.fromtimestamp(obj[0] + obj[1] / 1000000.)
    def _encode(self, obj, context, path):
        epoch = datetime.datetime.utcfromtimestamp(0)
        return [int((obj-epoch).total_seconds()), 0]

        # offset = time.mktime(*obj.timetuple())
        # sec = int(offset)
        # usec = (offset - sec) * 1000000
        # return (sec, usec)

packet = Struct(
    "time" / MicrosecAdapter(Int32ul >> Int32ul),
    "length" / Int32ul,
    Padding(4),
    "data" / Bytes(this.length),
)

cap_file = Padded(24, GreedyRange(packet))



"""
Portable Executable (PE) 32 bit, little endian
Used on MSWindows systems (including DOS) for EXEs and DLLs

1999 paper:
http://download.microsoft.com/download/1/6/1/161ba512-40e2-4cc9-843a-923143f3456c/pecoff.doc

2006 with updates relevant for .NET:
http://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/pecoff_v8.doc
"""

from construct import *
import time


# code uses linux epoch in microsoft format?
class UTCTimeStampAdapter(Adapter):
    def _decode(self, obj, context, path):
        return time.ctime(obj)
    def _encode(self, obj, context, path):
        return int(time.mktime(time.strptime(obj)))

# use core Timestamp
UTCTimeStamp = UTCTimeStampAdapter(Int32ul)


class NamedSequence(Adapter):
    """
    creates a mapping between the elements of a sequence and their respective
    names. this is useful for sequences of a variable length, where each
    element in the sequence has a name (as is the case with the data
    directories of the PE header)
    """
    __slots__ = ["mapping", "rev_mapping"]
    prefix = "unnamed_"

    def __init__(self, subcon, mapping):
        super(NamedSequence, self).__init__(subcon)
        self.mapping = mapping
        self.rev_mapping = dict((v, k) for k, v in mapping.items())

    def _decode(self, obj, context, path):
        obj2 = Container()
        for i, item in enumerate(obj):
            if i in self.mapping:
                name = self.mapping[i]
            else:
                name = "%s%d" % (self.prefix, i)
            setattr(obj2, name, item)
        return obj2

    def _encode(self, obj, context, path):
        obj2 = [None] * len(obj)
        for name, value in obj.items():
            if name in self.rev_mapping:
                index = self.rev_mapping[name]
            elif name.startswith("__"):
                obj2.pop(-1)
                continue
            elif name.startswith(self.prefix):
                index = int(name.split(self.prefix)[1])
            else:
                raise ValueError("no mapping defined for %r" % (name,))
            obj2[index] = value
        return obj2


msdos_header = Struct(
    Const(b"MZ"),
    "partPag" / Int16ul,
    "page_count" / Int16ul,
    "relocation_count" / Int16ul,
    "header_size" / Int16ul,
    "minmem" / Int16ul,
    "maxmem" / Int16ul,
    "relocation_stackseg" / Int16ul,
    "exe_stackptr" / Int16ul,
    "checksum" / Int16ul,
    "exe_ip" / Int16ul,
    "relocation_codeseg" / Int16ul,
    "table_offset" / Int16ul,
    "overlay" / Int16ul,
    Padding(8),
    "oem_id" / Int16ul,
    "oem_info" / Int16ul,
    Padding(20),
    "coff_header_pointer" / Int32ul,
    "_assembly_start" / Tell,
    "code" / Bytes(this.coff_header_pointer - this._assembly_start),
)

symbol_table = "symbol_table" / Struct(
    "name" / PaddedString(8, "utf8"),
    "value" / Int32ul,
    "section_number" / Enum(
        ExprAdapter(Int16sl, obj_-1, obj_+1),
        UNDEFINED = -1,
        ABSOLUTE = -2,
        DEBUG = -3,
    ),
    "complex_type" / Enum(Int8ul,
        NULL = 0,
        POINTER = 1,
        FUNCTION = 2,
        ARRAY = 3,
    ),
    "base_type" / Enum(Int8ul,
        NULL = 0,
        VOID = 1,
        CHAR = 2,
        SHORT = 3,
        INT = 4,
        LONG = 5,
        FLOAT = 6,
        DOUBLE = 7,
        STRUCT = 8,
        UNION = 9,
        ENUM = 10,
        MOE = 11,
        BYTE = 12,
        WORD = 13,
        UINT = 14,
        DWORD = 15,
    ),
    "storage_class" / Enum(Int8ul,
        NULL = 0,
        AUTOMATIC = 1,
        EXTERNAL = 2,
        STATIC = 3,
        REGISTER = 4,
        EXTERNAL_DEF = 5,
        LABEL = 6,
        UNDEFINED_LABEL = 7,
        MEMBER_OF_STRUCT = 8,
        ARGUMENT = 9,
        STRUCT_TAG = 10,
        MEMBER_OF_UNION = 11,
        UNION_TAG = 12,
        TYPE_DEFINITION = 13,
        UNDEFINED_STATIC = 14,
        ENUM_TAG = 15,
        MEMBER_OF_ENUM = 16,
        REGISTER_PARAM = 17,
        BIT_FIELD = 18,
        BLOCK = 100,
        FUNCTION = 101,
        END_OF_STRUCT = 102,
        FILE = 103,
        SECTION = 104,
        WEAK_EXTERNAL = 105,
        END_OF_FUNCTION = 255,
    ),
    "number_of_aux_symbols" / Int8ul,
    "aux_symbols" / Array(this.number_of_aux_symbols, Bytes(18))
)

coff_header = Struct(
    Const(b"PE\x00\x00"),
    "machine_type" / Enum(Int16ul,
        UNKNOWN = 0x0,
        AM33 = 0x1d3,
        AMD64 = 0x8664,
        ARM = 0x1c0,
        EBC = 0xebc,
        I386 = 0x14c,
        IA64 = 0x200,
        M32R = 0x9041,
        MIPS16 = 0x266,
        MIPSFPU = 0x366,
        MIPSFPU16 = 0x466,
        POWERPC = 0x1f0,
        POWERPCFP = 0x1f1,
        R4000 = 0x166,
        SH3 = 0x1a2,
        SH3DSP = 0x1a3,
        SH4 = 0x1a6,
        SH5= 0x1a8,
        THUMB = 0x1c2,
        WCEMIPSV2 = 0x169,
    ),
    "number_of_sections" / Int16ul,
    "time_stamp" / UTCTimeStamp,
    "symbol_table_pointer" / Int32ul,
    "number_of_symbols" / Int32ul,
    "optional_header_size" / Int16ul,
    "characteristics" / FlagsEnum(Int16ul,
        RELOCS_STRIPPED = 0x0001,
        EXECUTABLE_IMAGE = 0x0002,
        LINE_NUMS_STRIPPED = 0x0004,
        LOCAL_SYMS_STRIPPED = 0x0008,
        AGGRESSIVE_WS_TRIM = 0x0010,
        LARGE_ADDRESS_AWARE = 0x0020,
        MACHINE_16BIT = 0x0040,
        BYTES_REVERSED_LO = 0x0080,
        MACHINE_32BIT = 0x0100,
        DEBUG_STRIPPED = 0x0200,
        REMOVABLE_RUN_FROM_SWAP = 0x0400,
        SYSTEM = 0x1000,
        DLL = 0x2000,
        UNIPROCESSOR_ONLY = 0x4000,
        BIG_ENDIAN_MACHINE = 0x8000,
    ),
    "symbol_table" / Pointer(this.symbol_table_pointer, 
        Array(this.number_of_symbols, symbol_table))
)

PEPlusField = IfThenElse(this.pe_type == "PE32_plus", Int64ul, Int32ul)

optional_header = Struct(
    # standard fields
    "pe_type" / Enum(Int16ul,
        PE32 = 0x10b,
        PE32_plus = 0x20b,
    ),
    "major_linker_version" / Int8ul,
    "minor_linker_version" / Int8ul,
    "code_size" / Int32ul,
    "initialized_data_size" / Int32ul,
    "uninitialized_data_size" / Int32ul,
    "entry_point_pointer" / Int32ul,
    "base_of_code" / Int32ul,

    # only in PE32 files
    "base_of_data" / If(this.pe_type == "PE32", Int32ul),

    # WinNT-specific fields
    "image_base" / PEPlusField,
    "section_aligment" / Int32ul,
    "file_alignment" / Int32ul,
    "major_os_version" / Int16ul,
    "minor_os_version" / Int16ul,
    "major_image_version" / Int16ul,
    "minor_image_version" / Int16ul,
    "major_subsystem_version" / Int16ul,
    "minor_subsystem_version" / Int16ul,
    Padding(4),
    "image_size" / Int32ul,
    "headers_size" / Int32ul,
    "checksum" / Int32ul,
    "subsystem" / Enum(Int16ul,
        UNKNOWN = 0,
        NATIVE = 1,
        WINDOWS_GUI = 2,
        WINDOWS_CUI = 3,
        POSIX_CIU = 7,
        WINDOWS_CE_GUI = 9,
        EFI_APPLICATION = 10,
        EFI_BOOT_SERVICE_DRIVER = 11,
        EFI_RUNTIME_DRIVER = 12,
        EFI_ROM = 13,
        XBOX = 14,
    ),
    "dll_characteristics" / FlagsEnum(Int16ul,
        NO_BIND = 0x0800,
        WDM_DRIVER = 0x2000,
        TERMINAL_SERVER_AWARE = 0x8000,
    ),
    "reserved_stack_size" / PEPlusField,
    "stack_commit_size" / PEPlusField,
    "reserved_heap_size" / PEPlusField,
    "heap_commit_size" / PEPlusField,
    "loader_flags" / Int32ul,
    "number_of_data_directories" / Int32ul,

    "data_directories" / NamedSequence(
        Array(this.number_of_data_directories,
            Struct(
                "address" / Int32ul,
                "size" / Int32ul,
            )
        ),
        mapping = {
            0 : 'export_table',
            1 : 'import_table',
            2 : 'resource_table',
            3 : 'exception_table',
            4 : 'certificate_table',
            5 : 'base_relocation_table',
            6 : 'debug',
            7 : 'architecture',
            8 : 'global_ptr',
            9 : 'tls_table',
            10 : 'load_config_table',
            11 : 'bound_import',
            12 : 'import_address_table',
            13 : 'delay_import_descriptor',
            14 : 'complus_runtime_header',
        }
    ),
)

section = "section" / Struct(
    "name" / PaddedString(8, "utf8"),
    "virtual_size" / Int32ul,
    "virtual_address" / Int32ul,
    "raw_data_size" / Int32ul,
    "raw_data_pointer" / Int32ul,
    "relocations_pointer" / Int32ul,
    "line_numbers_pointer" / Int32ul,
    "number_of_relocations" / Int16ul,
    "number_of_line_numbers" / Int16ul,
    "characteristics" / FlagsEnum(Int32ul,
        TYPE_REG = 0x00000000,
        TYPE_DSECT = 0x00000001,
        TYPE_NOLOAD = 0x00000002,
        TYPE_GROUP = 0x00000004,
        TYPE_NO_PAD = 0x00000008,
        TYPE_COPY = 0x00000010,
        CNT_CODE = 0x00000020,
        CNT_INITIALIZED_DATA = 0x00000040,
        CNT_UNINITIALIZED_DATA = 0x00000080,
        LNK_OTHER = 0x00000100,
        LNK_INFO = 0x00000200,
        TYPE_OVER = 0x00000400,
        LNK_REMOVE = 0x00000800,
        LNK_COMDAT = 0x00001000,
        MEM_FARDATA = 0x00008000,
        MEM_PURGEABLE = 0x00020000,
        MEM_16BIT = 0x00020000,
        MEM_LOCKED = 0x00040000,
        MEM_PRELOAD = 0x00080000,
        ALIGN_1BYTES = 0x00100000,
        ALIGN_2BYTES = 0x00200000,
        ALIGN_4BYTES = 0x00300000,
        ALIGN_8BYTES = 0x00400000,
        ALIGN_16BYTES = 0x00500000,
        ALIGN_32BYTES = 0x00600000,
        ALIGN_64BYTES = 0x00700000,
        ALIGN_128BYTES = 0x00800000,
        ALIGN_256BYTES = 0x00900000,
        ALIGN_512BYTES = 0x00A00000,
        ALIGN_1024BYTES = 0x00B00000,
        ALIGN_2048BYTES = 0x00C00000,
        ALIGN_4096BYTES = 0x00D00000,
        ALIGN_8192BYTES = 0x00E00000,
        LNK_NRELOC_OVFL = 0x01000000,
        MEM_DISCARDABLE = 0x02000000,
        MEM_NOT_CACHED = 0x04000000,
        MEM_NOT_PAGED = 0x08000000,
        MEM_SHARED = 0x10000000,
        MEM_EXECUTE = 0x20000000,
        MEM_READ = 0x40000000,
        MEM_WRITE = 0x80000000,
    ),

    "raw_data" / Pointer(this.raw_data_pointer, 
        Bytes(this.raw_data_size)),

    "line_numbers" / Pointer(this.line_numbers_pointer,
        Array(this.number_of_line_numbers,
            Struct(
                "type" / Int32ul,
                "line_number" / Int16ul,
            )
        )
    ),

    "relocations" / Pointer(this.relocations_pointer,
        Array(this.number_of_relocations,
            Struct(
                "virtual_address" / Int32ul,
                "symbol_table_index" / Int32ul,
                "type" / Int16ul,
            )
        )
    ),
)

pe32_file = "pe32_file" / Struct(
    "msdos_header" / msdos_header,
    "coff_header" / coff_header,
    "_start_of_optional_header" / Tell,
    "optional_header" / optional_header,
    "_end_of_optional_header" / Tell,
    # this code is just weird
    Padding(lambda this: min(0, this.coff_header.optional_header_size - this._end_of_optional_header + this._start_of_optional_header)),
    "sections" / Array(this.coff_header.number_of_sections, 
        section),
)



from construct import *
from construct.lib import *
import sys, traceback, pdb, inspect


class Probe(Construct):
    r"""
    Probe that dumps the context, and some stream content (peeks into it) to the screen to aid the debugging process. It can optionally limit itself to a single context entry, instead of printing entire context.

    :param into: optional, None by default, or context lambda
    :param lookahead: optional, integer, number of bytes to dump from the stream

    Example::

        >>> d = Struct(
        ...     "count" / Byte,
        ...     "items" / Byte[this.count],
        ...     Probe(lookahead=32),
        ... )
        >>> d.parse(b"\x05abcde\x01\x02\x03")

        --------------------------------------------------
        Probe, path is (parsing), into is None
        Stream peek: (hexlified) b'010203'...
        Container: 
            count = 5
            items = ListContainer: 
                97
                98
                99
                100
                101
        --------------------------------------------------

    ::

        >>> d = Struct(
        ...     "count" / Byte,
        ...     "items" / Byte[this.count],
        ...     Probe(this.count),
        ... )
        >>> d.parse(b"\x05abcde\x01\x02\x03")

        --------------------------------------------------
        Probe, path is (parsing), into is this.count
        5
        --------------------------------------------------

    """

    def __init__(self, into=None, lookahead=None):
        super(Probe, self).__init__()
        self.flagbuildnone = True
        self.into = into
        self.lookahead = lookahead

    def _parse(self, stream, context, path):
        self.printout(stream, context, path)

    def _build(self, obj, stream, context, path):
        self.printout(stream, context, path)

    def _sizeof(self, context, path):
        self.printout(None, context, path)
        return 0

    def _emitparse(self, code):
        return "print(%s)" % (self.into,) if self.into else "print(this)"

    def printout(self, stream, context, path):
        print("--------------------------------------------------")
        print("Probe, path is %s, into is %r" % (path, self.into, ))

        if self.lookahead and stream is not None:
            fallback = stream.tell()
            datafollows = stream.read(self.lookahead)
            stream.seek(fallback)
            if datafollows:
                print("Stream peek: (hexlified) %s..." % (hexlify(datafollows), ))
            else:
                print("Stream peek: EOF reached")

        if context is not None:
            if self.into:
                try:
                    subcontext = self.into(context)
                    print(subcontext)
                except Exception:
                    print("Failed to compute %r on the context %r" % (self.into, context, ))
            else:
                print(context)
        print("--------------------------------------------------")


class Debugger(Subconstruct):
    r"""
    PDB-based debugger. When an exception occurs in the subcon, a debugger will appear and allow you to debug the error (and even fix it on-the-fly).
    
    :param subcon: Construct instance, subcon to debug
    
    Example::
    
        >>> Debugger(Byte[3]).build([])

        --------------------------------------------------
        Debugging exception of <Array: None>
        path is (building)
          File "/media/arkadiusz/MAIN/GitHub/construct/construct/debug.py", line 192, in _build
            return self.subcon._build(obj, stream, context, path)
          File "/media/arkadiusz/MAIN/GitHub/construct/construct/core.py", line 2149, in _build
            raise RangeError("expected %d elements, found %d" % (count, len(obj)))
        construct.core.RangeError: expected 3 elements, found 0

        > /media/arkadiusz/MAIN/GitHub/construct/construct/core.py(2149)_build()
        -> raise RangeError("expected %d elements, found %d" % (count, len(obj)))
        (Pdb) q
        --------------------------------------------------
    """

    def _parse(self, stream, context, path):
        try:
            return self.subcon._parse(stream, context, path)
        except Exception:
            self.retval = NotImplemented
            self.handle_exc(path, msg="(you can set self.retval, which will be returned from method)")
            if self.retval is NotImplemented:
                raise
            else:
                return self.retval

    def _build(self, obj, stream, context, path):
        try:
            return self.subcon._build(obj, stream, context, path)
        except Exception:
            self.handle_exc(path)

    def _sizeof(self, context, path):
        try:
            return self.subcon._sizeof(context, path)
        except Exception:
            self.handle_exc(path)

    def _emitparse(self, code):
        return self.subcon._compileparse(code)

    def handle_exc(self, path, msg=None):
        print("--------------------------------------------------")
        print("Debugging exception of %r" % (self.subcon, ))
        print("path is %s" % (path, ))
        print("".join(traceback.format_exception(*sys.exc_info())[1:]))
        if msg:
            print(msg)
        pdb.post_mortem(sys.exc_info()[2])
        print("--------------------------------------------------")



"""
Master Boot Record
The first sector on disk, contains the partition table, bootloader, et al.

http://www.win.tue.nl/~aeb/partitions/partition_types-1.html
"""

from construct import *

mbr_format = Struct(
    "bootloader_code" / Bytes(446),
    "partitions" / Array(4, Struct(
        "state" / Enum(Byte,
            INACTIVE = 0x00,
            ACTIVE = 0x80,
        ),
        "beginning" / BitStruct(
            "head" / BitsInteger(8),
            "sect" / BitsInteger(6),
            "cyl" / BitsInteger(10),
        ),
        "type" / Enum(Byte,
            Nothing = 0x00,
            FAT12 = 0x01,
            XENIX_ROOT = 0x02,
            XENIX_USR = 0x03,
            FAT16_old = 0x04,
            Extended_DOS = 0x05,
            FAT16 = 0x06,
            NTFS = 0x07,
            FAT32 = 0x0B,
            FAT32_LBA = 0x0C,
            ExtendedWithLBA = 0x0F,
            LINUX_SWAP = 0x82,
            LINUX_NATIVE = 0x83,
        ),
        "ending" / BitStruct(
            "head" / BitsInteger(8),
            "sect" / BitsInteger(6),
            "cyl" / BitsInteger(10),
        ),
        "sector_offset" / Int32ub, # offset from MBR in sectors
        "size" / Int32ub, # in sectors
    )),
    "signature" / Const(b"\x55\xAA"),
)



import pytest
xfail = pytest.mark.xfail
skip = pytest.mark.skip
skipif = pytest.mark.skipif

import os, math, random, collections, itertools, io, hashlib, binascii, tempfile

from construct import *
from construct.lib import *

ident = lambda x: x
devzero = open("/dev/zero", "rb")


def raises(func, *args, **kw):
    try:
        return func(*args, **kw)
    except Exception as e:
        return e.__class__


def common(format, datasample, objsample, sizesample=SizeofError, **kw):
    obj = format.parse(datasample, **kw)
    assert obj == objsample
    data = format.build(objsample, **kw)
    assert data == datasample
    # following are implied by above (re-parse and re-build)
    # assert format.parse(format.build(obj)) == obj
    # assert format.build(format.parse(data)) == data
    if isinstance(sizesample, int):
        size = format.sizeof(**kw)
        assert size == sizesample
    else:
        size = raises(format.sizeof, **kw)
        assert size == sizesample


def commonhex(format, hexdata):
    commonbytes(format, binascii.unhexlify(hexdata))


def commondumpdeprecated(format, filename):
    filename = "tests/deprecated_gallery/blobs/" + filename
    with open(filename,'rb') as f:
        data = f.read()
    commonbytes(format, data)


def commondump(format, filename):
    filename = "tests/gallery/blobs/" + filename
    with open(filename,'rb') as f:
        data = f.read()
    commonbytes(format, data)


def commonbytes(format, data):
    obj = format.parse(data)
    data2 = format.build(obj)



"""
Extension 2 (ext2) used in Linux systems
"""

from construct import *

Char = SLInt8
UChar = ULInt8
Short = SLInt16
UShort = ULInt16
Long = SLInt32
ULong = ULInt32

BlockPointer = Struct(
    "block_number" / ULong,
    # WARNING: unnamed field?
    OnDemandPointer(this.block_number),
)

superblock = Struct(
    "inodes_count" / ULong,
    "blocks_count" / ULong,
    "reserved_blocks_count" / ULong,
    "free_blocks_count" / ULong,
    "free_inodes_count" / ULong,
    "first_data_block" / ULong,
    "log_block_size" / Enum(ULong, 
        OneKB = 0,
        TwoKB = 1,
        FourKB = 2,
    ),
    "log_frag_size" / Long,
    "blocks_per_group" / ULong,
    "frags_per_group" / ULong,
    "inodes_per_group" / ULong,
    "mtime" / ULong,
    "wtime" / ULong,
    "mnt_count" / UShort,
    "max_mnt_count" / Short,
    "magic" / Const(UShort, 0xEF53),
    "state" / UShort,
    "errors" / UShort,
    Padding(2),
    "lastcheck" / ULong,
    "checkinterval" / ULong,
    "creator_os" / ULong,
    "rev_level" / ULong,
    Padding(235*4),
)

group_descriptor = Struct(
    "block_bitmap" / ULong,
    "inode_bitmap" / ULong,
    "inode_table" / ULong,
    "free_blocks_count" / UShort,
    "free_inodes_count" / UShort,
    "used_dirs_count" / UShort,
    Padding(14),
)

inode = Struct(
    "mode" / FlagsEnum(UShort,
        IXOTH = 0x0001,
        IWOTH = 0x0002,
        IROTH = 0x0004,
        IRWXO = 0x0007,
        IXGRP = 0x0008,
        IWGRP = 0x0010,
        IRGRP = 0x0020,
        IRWXG = 0x0038,
        IXUSR = 0x0040,
        IWUSR = 0x0080,
        IRUSR = 0x0100,
        IRWXU = 0x01C0,
        ISVTX = 0x0200,
        ISGID = 0x0400,
        ISUID = 0x0800,
        IFIFO = 0x1000,
        IFCHR = 0x2000,
        IFDIR = 0x4000,
        IFBLK = 0x6000,
        IFREG = 0x8000,
        IFLNK = 0xC000,
        IFSOCK = 0xA000,
        IFMT = 0xF000,
    ),
    "uid" / UShort,
    "size" / ULong,
    "atime" / ULong,
    "ctime" / ULong,
    "mtime" / ULong,
    "dtime" / ULong,
    "gid" / UShort,
    "links_count" / UShort,
    "blocks" / ULong,
    "flags" / FlagsEnum(ULong,
        SecureDelete = 0x0001,
        AllowUndelete = 0x0002,
        Compressed = 0x0004,
        Synchronous = 0x0008,
    ),
    Padding(4),
    # WARNING: doubled name
    "blocks" / ULong[12],
    "indirect1_block" / ULong,
    "indirect2_block" / ULong,
    "indirect3_block" / ULong,
    "version" / ULong,
    "file_acl" / ULong,
    "dir_acl" / ULong,
    "faddr" / ULong,
    "frag" / UChar,
    "fsize" / UChar,
    Padding(10),
)

# special inodes
EXT2_BAD_INO = 1
EXT2_ROOT_INO = 2
EXT2_ACL_IDX_INO = 3
EXT2_ACL_DATA_INO = 4
EXT2_BOOT_LOADER_INO = 5
EXT2_UNDEL_DIR_INO = 6
EXT2_FIRST_INO = 11 

directory_record = Struct(
    "inode" / ULong,
    "rec_length" / UShort,
    "name_length" / UShort,
    "name" / Bytes(this.name_length),
    Padding(this.rec_length - this.name_length),
)



from declarativeunittest import *
from construct.lib.binary import *


def test_integer2bits():
    assert integer2bits(0, 0) == b""
    assert integer2bits(19, 5) == b"\x01\x00\x00\x01\x01"
    assert integer2bits(19, 8) == b'\x00\x00\x00\x01\x00\x00\x01\x01'
    assert integer2bits(19, 3) == b'\x00\x01\x01'
    assert integer2bits(-13, 5) == b"\x01\x00\x00\x01\x01"
    assert integer2bits(-13, 8) == b"\x01\x01\x01\x01\x00\x00\x01\x01"
    assert raises(integer2bits, 19, -1) == ValueError
    assert raises(integer2bits, -19, -1) == ValueError

def test_integer2bytes():
    assert integer2bytes(0, 0) == b""
    assert integer2bytes(0, 4) == b"\x00\x00\x00\x00"
    assert integer2bytes(1, 4) == b"\x00\x00\x00\x01"
    assert integer2bytes(255, 4) == b"\x00\x00\x00\xff"
    assert integer2bytes(-1, 4) == b"\xff\xff\xff\xff"
    assert integer2bytes(-255, 4) == b"\xff\xff\xff\x01"
    assert raises(integer2bytes, 19, -1) == ValueError
    assert raises(integer2bytes, -19, -1) == ValueError

def test_bits2integer():
    assert bits2integer(b"\x01\x00\x00\x01\x01") == 19
    assert bits2integer(b"\x01\x00\x00\x01\x01", True) == -13

def test_cross_integers():
    for i in [-300,-255,-100,-1,0,1,100,255,300]:
        assert bits2integer(integer2bits(i,64),signed=(i<0)) == i
        assert bytes2integer(integer2bytes(i,8),signed=(i<0)) == i
        assert bits2bytes(integer2bits(i,64)) == integer2bytes(i,8)
        assert bytes2bits(integer2bytes(i,8)) == integer2bits(i,64)

def test_bytes2bits():
    assert bytes2bits(b"ab") == b"\x00\x01\x01\x00\x00\x00\x00\x01\x00\x01\x01\x00\x00\x00\x01\x00"
    assert bytes2bits(b"") == b""

def test_bits2bytes():
    assert bits2bytes(b"\x00\x01\x01\x00\x00\x00\x00\x01\x00\x01\x01\x00\x00\x00\x01\x00") == b"ab"
    assert bits2bytes(b"") == b""
    assert raises(bits2bytes, b"\x00") == ValueError
    assert raises(bits2bytes, b"\x00\x00\x00\x00\x00\x00\x00") == ValueError

def test_swapbytes():
    assert swapbytes(b"") == b""
    assert swapbytes(b"abcd") == b"dcba"

def test_swapbytesinbits():
    assert swapbytesinbits(b"0000000011111111") == b"1111111100000000"
    assert swapbytesinbits(b"") == b""
    assert raises(swapbytesinbits, b"1") == ValueError

def test_swapbitsinbytes():
    assert swapbitsinbytes(b'') == b''
    assert swapbitsinbytes(b'\xf0') == b'\x0f'
    assert swapbitsinbytes(b'\xf0\x00') == b'\x0f\x00'



from construct import *

d = Struct(
    "count" / Int32ul,
    "items" / Array(this.count, Struct(
        "num1" / Int8ul,
        "num2" / Int24ul,
        "flags" / BitStruct(
            "bool1" / Flag,
            "num4" / BitsInteger(3),
            Padding(4),
        ),
        "fixedarray1" / Array(3, Int8ul),
        "name1" / CString("utf8"),
        "name2" / PascalString(Int8ul, "utf8"),
    )),
)


data = d.build(dict(count=1000, items=[dict(num1=0, num2=0, flags=dict(bool1=True, num4=0), fixedarray1=[0,0,0], name1=u"...", name2=u"...") for i in range(1000)]))
with open("blob","wb") as f:
    f.write(data)

# from timeit import timeit
# d.parse(data)
# parsetime = timeit(lambda: d.parse(data), number=1000)/1000
# print("Timeit measurements:")
# print("parsing:           {:.10f} sec/call".format(parsetime))

d = d.compile()
print(d.benchmark(data))



from declarativeunittest import *
from construct.lib.bitstream import *


def test_restreamed():
    # tested by Bitwise Bytewise ByteSwapped BitsSwapped cases
    pass

def test_rebuffered():
    z = b"0"

    print("sequential read")
    bstream = RebufferedBytesIO(io.BytesIO(z*1000))
    assert bstream.read(1000) == z*1000

    print("random reads")
    data = os.urandom(1000)
    bstream = RebufferedBytesIO(io.BytesIO(data))
    for i in range(50):
        o1 = random.randrange(0, 480)
        o2 = random.randrange(520, 1000)
        assert bstream.seek(o1) == o1
        assert bstream.tell() == o1
        assert bstream.read(o2-o1) == data[o1:o2]
        assert bstream.tell() == o2

    print("sequential writes")
    bstream = RebufferedBytesIO(io.BytesIO())
    for i in range(10):
        assert bstream.write(z*100) == 100
    assert bstream.seek(0) == 0
    assert bstream.read(1000) == z*1000

    print("random writes")
    data = os.urandom(1000)
    bstream = RebufferedBytesIO(io.BytesIO())
    assert bstream.write(data) == len(data)
    for i in range(50):
        o1 = random.randrange(0, 480)
        o2 = random.randrange(520, 1000)
        assert bstream.seek(o1) == o1
        assert bstream.tell() == o1
        assert bstream.write(data[o1:o2]) == o2-o1
        assert bstream.tell() == o2
    assert bstream.seek(0) == 0
    assert bstream.read(len(data)) == data

    print("cutting off trail")
    data = os.urandom(1000)
    bstream = RebufferedBytesIO(io.BytesIO(data), tailcutoff=50)
    for i in range(15):
        at = bstream.tell()
        assert bstream.read(50) == data[at:at+50]
        jumpback = random.randrange(1, 19)
        assert bstream.seek(-jumpback, 1)
